clear; close all; clc;

% number of repeated runs for Monte-Carlo simulation
NUMBER_OF_TESTS = 100;

% length of an episode, i.e. number of sensing operations (sensing budget)
MAX_SENSING_STEPS = 1000;

% Uncomment to get repeatable runs
RandStream.setDefaultStream(RandStream('mt19937ar','seed',45));

global tree;
global alpha;
global beta;
global lower_thresh;
global upper_thresh;
global treeFigureHandler;
global epsilon;
global pnull; % probability of the hypothesis target not present

global maxDepth;

global xslices;
global yslices;

global cost_ascend; % cost increase for ascending
global bonus_descend; % bonus for descending

edge = 256;

% for the time being disregard ascending/descending
cost_ascend = 1000;
bonus_descend = 0;

% Define constants
xslices = edge;
yslices = edge;

% threshold to establish when two reals are the same
epsilon = 0.00001;

maxDepth = 8+1;     % recall root node starts at depth=1

% sensor perfomance
x = 1:10;
% these are the alpha values parametrized by depth
alpha = 0.2 - 0.01*x;
% these are the beta values parametrized by depth
beta = 0.3 - 0.02*x;

% Lower search decision thresholds for clearing a cell
lower_thresh = 0.0 + 0.01*x;

% Upper search decision thresholds for designating a cell as occupied
upper_thresh = 0.81 + 0.01*x;

% Define belief evolution vector
%  - initialize aggregate belief
initProb = 0.5;
belief_evol = [initProb];

% target position  (x,y) positions in columns
% in a type one PQ there is always only one intruder
% target = [ 200  200]';
target = [ 200.5  200.5]';      % remember target must be at deepest resolution

% threshold to establish when two reals are the same
epsilon = 0.00001;

tree = CreateNode(0,initProb,0);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Here one decides how to initialize the hierarchical data structure
% See the different alternatives in terms of size
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% % this one will actually create a tree with 297 nodes
% BuildTreeMaxNodes(300,'init_belief_map_256.mat');

% this one will create a tree with 341 nodes (uniform depth)
%BuildTreeMaxDepth(5,'init_belief_map_256.mat');

% this one will create a tree with 85 nodes (uniform depth)
% BuildTreeMaxDepth(4,'init_belief_map_256.mat');

% this one will actually create a tree with 297 nodes
BuildTreeMaxNodes(150,'init_belief_map_256.mat');

saveTree = tree;

% stores 1/0 if the target was found in the i-th run
success = zeros(1,NUMBER_OF_TESTS);
% stores the number of steps needed to complete the i-th run
time_steps = zeros(1,NUMBER_OF_TESTS);
% stores the target position found by the search algorithm
final_positions = zeros(2,NUMBER_OF_TESTS);
% at position (i,j) stores the depth of the sensor in the j-th step of the
% i-th run
depths = zeros(NUMBER_OF_TESTS,MAX_SENSING_STEPS);
% at position (i,j) stores the number of nodes in the tree in the j-th step
% of the i-th run
tree_size = zeros(NUMBER_OF_TESTS,MAX_SENSING_STEPS);
% at position (i,j) stores the number of leaves in the tree in the j-th step
% of the i-th run
number_of_leaves = zeros(NUMBER_OF_TESTS,MAX_SENSING_STEPS);
% at position (i,j) stores the number of cleared nodes in the tree in the
% j-th step of the i-th run
number_of_cleared = zeros(NUMBER_OF_TESTS,MAX_SENSING_STEPS);


% Visualization flags
FLAG_SHOW_MAP = false;
FLAG_SHOW_QUAD = false;



for N = 1:NUMBER_OF_TESTS
    
    N
    
    tree = saveTree;
    
    
    if(FLAG_SHOW_MAP)
        gcmesh = figure;
        PlotEstimate(gcmesh);
    end
    
    if (FLAG_SHOW_QUAD)
        treeFigureHandler = DrawQuadTree;
        
        figure(treeFigureHandler); hold on;
        
        h_sensor(1) = plot3(-100, -100, max(get(gca, 'ZLim')), 'ro', 'MarkerSize', 10, 'MarkerFaceColor', 'r' );  % searcher position
        h_sensor(2) = plot3(-100, -100, max(get(gca, 'ZLim')), 'rx', 'MarkerSize', 10 );                          % searcher next position
        hold off;
    end
    
    cleared_cells = [];
    
    pnull = 1-tree(1).p; % probability the target is outside
    
    c_curr = 1; % position in the tree where the sensor is currently located;
    % we assume in the beginning it is at the center, maximum height, i.e. root
    % of the tree
    
    c_next = 1;
    c_goal = 2;
    
    % Define belief trajectory vector
    %  - initialize with start location
    belief_traj = c_curr;
    
    curPath = [];
    
    %=============================================
    % SIMULATION
    %  - SENSE-->PLAN-->ACT flow
    %=============================================
    for i = 1:MAX_SENSING_STEPS
        % Time stamp
        %        fprintf('--> Time step %4d\n', i);
        
        
        % record relevant information
        tree_size(N,i) = numel(tree);
        number_of_leaves(N,i) = sum(cell2mat({tree(:).leaf}));
        number_of_cleared(N,i) = sum(cell2mat({tree(:).cleared})); 
        depths(N,i) = tree(c_next).depth;
        
         %if there is no currently planned path need to make one
    if isempty(curPath)     
        %compute a new goal
        c_goal = GetMaximumPosition(c_curr);
        curPath = ComputeShortestPath(c_curr,c_goal);
    end
    
        %=========================
        % Searcher Information Gathering
        %  - observe desired location (if appropriate)
        %=========================
        
        % Determine if the sensor is on, dictated by whether at goal location
        %  - we don't (yet) sense while in transit
        if ( curPath(1) == c_goal )       % arrived at goal
            
            % read the sensor, and if necessary expands the tree
            [z,pos] = GetSensorReading(tree(c_goal).xc, tree(c_goal).yc, target);
            c_curr = pos;
 
            
            UpdatePosterior(z,pos);
            
            % Given current observation and updated map, define decision
            [isSearchDone, cleared_cells] = EvaluateSearchDecision( z, c_curr, cleared_cells );
            if (isSearchDone)
                %    disp('Search has been completed');
                time_steps(N) = i;
                if norm(([tree(c_curr).xc tree(c_curr).yc]' - target)) < 0.01
                    success(N) = 1;
                end
                final_positions(:,N) = [tree(c_curr).xc tree(c_curr).yc]';
                break;
            end;
             curPath = [];
        else
            % Update to next position en route to goal
            curPath(1) = [];
        c_curr = curPath(1);
        end          % end if(check for transit)
        
        
        %     fprintf('  main.m: c_curr=%d, c_goal=%d, c_next=%d\n', c_curr, c_goal, c_next);   
        
        %=========================
        % Data logging
        %=========================
        % Augment belief evolution vector
        belief_evol = [belief_evol; tree(1).p];
        
        % Record searcher trajectory
        belief_traj = [belief_traj; c_curr];
        
        
        %=========================
        % Visualization handling
        %=========================
        if ( FLAG_SHOW_MAP )
           % if mod(i,10) == 0 % plot once in a while to save time...
                PlotEstimate(gcmesh);
           % end
            
            % Annotating desired search location on probability map
            %  - need to recreate object since probability map figure is redrawn
            figure(gcmesh); hold on;
            h_goal = plot3(tree(c_goal).xc, tree(c_goal).yc, max(get(gca, 'ZLim')), 'rx', 'MarkerSize', 10 );
            hold off;
        end         % end if(check FLAG_SHOW_MAP flag)
        
        
        if (FLAG_SHOW_QUAD)
            % Annotating desired search location on quad tree diagram
            %  - don't need to recreate object since quad tree figure is NOT redrawn
            set( h_sensor(1), 'XData', tree(c_curr).xc, 'YData', tree(c_curr).yc);
            set( h_sensor(2), 'XData', tree(c_goal).xc, 'YData', tree(c_goal).yc);
            
            % Annotate any cleared cells on quadtree diagram
            %  - leave all annotations on figure
            if ( ~isempty(cleared_cells) )
                figure(treeFigureHandler); hold on;
                plot3( tree(cleared_cells(end)).xc, tree(cleared_cells(end)).yc, max(get(gca, 'ZLim')), 'ko', 'MarkerSize', 6 );
                hold off;
            end
            
        end
        
    end
    
    
end

save HierarchicalSearch